#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  player.py
#  
#  Copyright 2019 Stefan Born <born@math.tu-berlin.de>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
import time
import numpy as np

class DenseLayer:

    def __init__(self, n_ein, n_aus):
        self.weights = np.random.randn(n_aus, n_ein)*np.sqrt(2/n_ein)
        self.bias = np.zeros(n_aus)
        self.shape = (n_aus, n_ein)

    def calc(self, ein):
        return self.weights.dot(ein)+self.bias


class SigmaActivation:

    def calc(self, x):
        return np.exp(x)/(1+np.exp(x))


class ReLUActivation:

    def calc(self, x):
        return np.where(x > 0, x, 0)


class TanhActivation:

    def __init__(self, factor=1):
        self.factor = factor
        
    def calc(self, x):
        try:
            return self.factor*(np.exp(2*x)-1)/(np.exp(2*x)+1)
        except:
            if x>10:
                return self.factor
            else:
                return -self.factor
       

class SequentialModel:
    '''Diese Klasse erlaubt, die oben definierten Layers
    aneinanderzuhaengen und merkt  sich dabei Referenzen auf
    die jeweiligen Parameter. Man muss sich die Klasse als
    ein Steckbrett vorstellen. man stöpselt jeweils die
    Ausgänge der letzten Schicht in den Eingang der naechsten.
    Sehr primitiv -- wer sie verwendet, muss selbst aufpassen,
    dass die Dimensionen jeweils gleich sind.
    '''

    def __init__(self):
        self.layers = []
        self.params = []

    def add_layer(self, layer):
        self.layers.append(layer)
        try:
            self.params.append(layer.weights)
            self.params.append(layer.bias)
        except:
            pass

    def calc(self, x):
        for layer in self.layers:
            x = layer.calc(x)
        return x

    def recommend(self, x):
        val = self.calc(x)
        print("Start! ",val, "  Pause!")
        if np.random.rand() < val:
            return 0
        else:
            return 1

    def mutate_params(self, sigma=0.1):
        for p in self.params:
            a = sigma*np.random.randn(*p.shape) #------------
            #print("Parameterchange:",a) #------------
            p[...] += a #------------

    def copy_params(self, params):
        for p, pneu in zip(self.params, params):
            p[...] = pneu
            
# Hier wird ein Modell vorgeschlagen dass 10 lineare Funktionen
# auf den vier Eingangsvariablen bserechnet, dann eine nichtlineare
# Funktion anwendet, dann daraus eine lineare Funktion berechnet,
# dann die logistische Funktion anwendet, so dass als Wahrscheinlichkeiten
# interpretierbare Werte zwischen 0 und 1 herauskommen.


def make_model(ein=16, hidden=40, aus=2):

    model = SequentialModel()
    model.add_layer(DenseLayer(ein, hidden))
    model.add_layer(ReLUActivation())
    model.add_layer(DenseLayer(hidden, aus))
    model.add_layer(TanhActivation())
    return model      
